"""
Created on Fri Jul 19 2019
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from os import path
from matplotlib.colors import LinearSegmentedColormap, Normalize
import pkg_resources
# DEFAULT_FILE_PATH = path.join(path.dirname(__file__), 'ColTabs')

class OpendtectColormaps(object):
    def __init__(self):
        self.od_cm_file = pkg_resources.resource_filename('seisvis.opendtect_colormaps', 'ColTabs')
        self.color_dicts = None
        self.normalizer = Normalize(vmin=0, vmax=255)
        self._convert_to_cm()

    @property
    def cmaps(self):
        return list(self.color_dicts.keys())

    def __call__(self, cm):
        n_levels = self.color_dicts[cm]['N']
        n_levels = 256 if (n_levels is None or n_levels == 0) else n_levels
        cmap = LinearSegmentedColormap.from_list(
            cm,
            colors=self.color_dicts[cm]['cdict'],
            N=n_levels,
            gamma=1.0)
        undef_color = self.color_dicts[cm]['undef']
        if undef_color is not None:
            cmap.set_bad(undef_color)
        return cmap

    def _convert_to_cm(self):
        with open(self.od_cm_file, 'r') as fl:
            text = fl.readlines()

        text_select = []
        num = -1
        for line in text:
            if line.split('.')[0].isdigit():
                num_color = int(line.split('.')[0])-1
                if num_color != num:
                    num = num_color
                    text_select.append([])
                text_select[num].append(line.split(':'))

        text_select = [dict(a) for a in text_select]

        color_dicts = {}
        for dic in text_select:
            # print(self.get_name(dic))
            color_dicts[self.get_name(dic)] = {
                "cdict": self.get_cdict(dic),
                "undef": self.get_undef_color(dic),
                "N": self.get_nr_segments(dic)
            }
        self.color_dicts = color_dicts

    @staticmethod
    def get_name(abc):
        colormap_name = None
        for k in abc.keys():
            if "Name" in k:
                colormap_name = abc[k].rstrip('\n').lstrip(' ').\
                    replace('-', '_').replace(' ', '_')
                break
        return colormap_name

    def get_cdict(self, abc):
        cdict = []
        for k in abc.keys():
            if "Value-Color" in k:
                str_list = abc[k].rstrip("\n").split('`')
                str_list = str_list[:-1] if len(str_list) == 5 else str_list
                v = float(str_list[0])
                list_int = [self.normalizer(int(i)) for i in str_list[1:]]
                cdict.append(tuple((v, tuple(list_int))))
        return cdict

    def get_undef_color(self, abc):
        undef_color = None
        for k in abc.keys():
            if "Undef color" in k:
                str_list = abc[k].rstrip("\n").split('`')
                str_list = str_list[:-1] if len(str_list) == 4 else str_list
                list_int = [self.normalizer(int(i)) for i in str_list]
                undef_color = tuple(list_int)
                break
        return undef_color

    @staticmethod
    def get_nr_segments(abc):
        nr_segments = None
        for k in abc.keys():
            if "Nr segments" in k:
                nr_segments = int(abc[k].rstrip("\n"))
                break
        return nr_segments
    
    def print_cmap_names(self):
        print("Available color maps:")
        for name in self.cmaps:
            print("-", name)



# Copyright (c) 2025 Qi Pang.
# All rights reserved.

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize
from typing import Optional, Tuple, List, Union, Dict
import warnings
import itertools

from .plot_config import PlotConfig
from .data_cube import DataCube

ClipType = Union[str, Tuple[float, float], None]
SeismicType = Dict[str, Union[str, ClipType]]
#  {'type': 'deg10','cmap': 'gray','clip': None}  # or 'rubust' or (min, max)


class Seis3DPlotter:
    """3D data plotter"""
    def __init__(self, data_cube: Union[DataCube],size=None,
                       config: Optional[PlotConfig] = None):
        """
        size: [il_start, il_end, xl_start, xl_end, t_end, t_start]
        layers: list of str, each in format 'type:name'
        """
        self.config = config or PlotConfig()
        self.config.apply_matplotlib_settings()
        
        self.cube = data_cube
        self.size = size
    
    @property
    def shape(self):
        if self.size is None:
            raise ValueError("shape requires 3D data. 'size' is not defined.")
        return [
            self.size[1] - self.size[0] + 1,
            self.size[3] - self.size[2] + 1,
            self.size[4] - self.size[5] + 1,
        ]
    
    def _validate_array(self, array: np.ndarray) -> np.ndarray:
        """Validate and preprocess input array"""
        if not isinstance(array, np.ndarray):
            array = np.array(array)

        if array.ndim != 3:
            raise ValueError(f"Input array must be 3-dimensional, got {array.ndim}D")

        if array.size == 0:
            raise ValueError("Input array cannot be empty")

        # Handle NaN values
        if np.isnan(array).any():
            warnings.warn("Array contains NaN values, replacing with array mean")
            array = np.nan_to_num(array, nan=np.nanmean(array))

        return array

    def _calculate_optimal_aspect(self, shape: Tuple[int, int, int], scale_cons=1.2):
        """Calculate optimal aspect ratio"""
        nx, ny, nz = shape
        max_index = shape.index(max(shape))
        if max_index==2:
            ar_z = 1.0
            ar_x = (nz/nx)*scale_cons
            ar_y = (nz/nx)*scale_cons
        elif max_index==0:
            ar_z = (nx/nz)
            ar_x = scale_cons
            ar_y = (nx/ny)*scale_cons
        else:
            ar_z = (ny/nz)
            ar_y = scale_cons
            ar_x = (ny/nx)*scale_cons
        return (ar_x, ar_y, ar_z)

    def _calculate_optimal_position(self, shape: Tuple[int, int, int],
                                  pos: Optional[List[int]] = None) -> List[int]:
        """Calculate optimal slice positions using golden ratio"""
        nx, ny, nz = shape

        # Use golden ratio for optimal visual positioning
        golden_ratio = 0.618
        pos = [
            int(nx * golden_ratio)+self.size[0],
            int(ny * golden_ratio)+self.size[2],
            int(nz * 0.4)+self.size[5]  # Z-axis often represents time, choose shallower position
        ]

        return pos

    def _get_clip_values(self, array: np.ndarray,
                        clip: Optional[Union[Tuple[float, float], str]] = None) -> Tuple[float, float]:
        """Calculate data clipping range with various methods"""
        if clip is None:
            return array.min(), array.max()
        elif isinstance(clip, str):
            if clip == 'robust':
                # Robust clipping using IQR
                q25, q75 = np.percentile(array, [25, 75])
                iqr = q75 - q25
                return q25 - 1.5*iqr, q75 + 1.5*iqr
            else:
                raise ValueError(f"Unsupported clip type: {clip}")
        else:
            return float(clip[0]), float(clip[1])

    def _plot_quadrants(self, ax, array: np.ndarray, fixed_coord: str,
                       pos: List[int], cmap, vmin: float, vmax: float):
        """Plot quadrant slices"""
        # pos = pos.copy()
        nx, ny, nz = array.shape
        pos[0] = pos[0]-self.size[0]
        pos[1] = pos[1]-self.size[2]
        pos[2] = pos[2]-self.size[5]
        
        # Select slice
        slice_index = {
            'x': pos[0],
            'y': pos[1],
            'z': pos[2]
        }[fixed_coord]

        index = {
            'x': (slice_index, slice(None), slice(None)),
            'y': (slice(None), slice_index, slice(None)),
            'z': (slice(None), slice(None), slice_index),
        }[fixed_coord]

        plane_data = array[index]

        # Define quadrant splits
        if fixed_coord == 'x':
            quadrants = [
                plane_data[:pos[1], :pos[2]],
                plane_data[:pos[1], pos[2]:],
                plane_data[pos[1]:, :pos[2]],
                plane_data[pos[1]:, pos[2]:]
            ]
            splits = [(0 + self.size[2], pos[1] + self.size[2], ny + self.size[2]), 
                      (0 + self.size[5], pos[2] + self.size[5], nz + self.size[5])]
        elif fixed_coord == 'y':
            quadrants = [
                plane_data[:pos[0], :pos[2]],
                plane_data[:pos[0], pos[2]:],
                plane_data[pos[0]:, :pos[2]],
                plane_data[pos[0]:, pos[2]:]
            ]
            splits = [(0 + self.size[0], pos[0] + self.size[0], nx + self.size[0]), 
                      (0 + self.size[5], pos[2] + self.size[5], nz + self.size[5])]
        elif fixed_coord == 'z':
            quadrants = [
                plane_data[:pos[0], :pos[1]],
                plane_data[:pos[0], pos[1]:],
                plane_data[pos[0]:, :pos[1]],
                plane_data[pos[0]:, pos[1]:]
            ]
            splits = [(0 + self.size[0], pos[0] + self.size[0], nx + self.size[0]), 
                      (0 + self.size[2], pos[1] + self.size[2], ny + self.size[2])]

        # Plot quadrants
        for i, quadrant in enumerate(quadrants):
            if quadrant.size == 0:
                continue

            # Enhanced color mapping
            normalized_data = (quadrant - vmin) / (vmax - vmin)
            # normalized_data = np.clip(normalized_data, 0, 1)  # Ensure valid range
            facecolors = plt.get_cmap(cmap)(normalized_data)
            facecolors[..., 3] = 0.3 # Set transparency

            # Calculate grid coordinates
            if fixed_coord == 'x':
                y_start, y_end = splits[0][(i // 2)], splits[0][(i // 2) + 1]
                z_start, z_end = splits[1][(i % 2)], splits[1][(i % 2) + 1]
                
                
                Y, Z = np.mgrid[y_start:y_end, z_start:z_end]
                X = slice_index * np.ones_like(Y) + self.size[0]
            elif fixed_coord == 'y':
                x_start, x_end = splits[0][(i // 2)], splits[0][(i // 2) + 1]
                z_start, z_end = splits[1][(i % 2)], splits[1][(i % 2) + 1]
                
                
                X, Z = np.mgrid[x_start:x_end, z_start:z_end]
                Y = slice_index * np.ones_like(X) + self.size[2]
            elif fixed_coord == 'z':
                x_start, x_end = splits[0][(i // 2)], splits[0][(i // 2) + 1]
                y_start, y_end = splits[1][(i % 2)], splits[1][(i % 2) + 1]
                
                
                X, Y = np.mgrid[x_start:x_end, y_start:y_end]
                Z = slice_index * np.ones_like(X) + self.size[5]

            ax.plot_surface(X, Y, Z, rstride=1, cstride=1,
                          facecolors=facecolors, shade=False, alpha=1)

    def plot_slices(self, show_seismic_type: Optional[SeismicType] = None,
               show_properties_type: Optional[SeismicType] = None,
               show_horizons_type: Optional[SeismicType] = None,
               show_wells_type: Optional[SeismicType] = None,
               figsize: Tuple[int, int] = (10, 8),
               pos: Optional[List[int]] = None,
               view_pos: Optional[Tuple[int, int]] = None,
               labels: Optional[Tuple[str, str, str]] = None,
               title: Optional[str] = None,
               unit_label = None,
               lighting: bool = True,
               save_path: Optional[str] = None) -> Union[Tuple[plt.Figure, plt.Axes],
                                                  Tuple[plt.Figure, plt.Axes, dict]]:
                                                         
        "Orthogonal plane visualization of 3D seismic data at any location"
        if show_seismic_type is not None:
            array = self.cube.get('seismic', show_seismic_type['type'])
            older_clip = self._get_clip_values(array, show_seismic_type['clip'])
            cmap = self.config.get_cmap(show_seismic_type['cmap'])
        
        if show_properties_type is not None:
            array = self.cube.get('properties', show_properties_type['type'])
            older_clip = self._get_clip_values(array, show_properties_type['clip'])
            cmap = self.config.get_cmap(show_properties_type['cmap'])
            
        
        # Validate input
        array = self._validate_array(array)

        # Set defaults
        if cmap is None:
            cmap = self.config.default_cmap
            
        if view_pos is None:
            view_pos = (self.config.default_elevation, self.config.default_azimuth)
            
        if labels is None:
            labels = ('Inline', 'Xline', 'Time (ms)')
            
        # Calculate positions and clipping values
        if pos is None:
            pos = self._calculate_optimal_position(array.shape, pos)

        # Create figure with enhanced styling
        fig = plt.figure(figsize=figsize, facecolor=self.config.background_color)
        ax = fig.add_subplot(111, projection='3d')

        # Set aspect ratio for better 3D perception
        aspect_ratio = self._calculate_optimal_aspect(array.shape)
        ax.set_box_aspect((array.shape[0] * aspect_ratio[0],
                          array.shape[1] * aspect_ratio[1],
                          array.shape[2] * aspect_ratio[2]))

        
        # Plot three intersecting planes
        for coord in ['x', 'y', 'z']:
            self._plot_quadrants(ax, array, coord, pos=pos.copy(), cmap=cmap, 
                                  vmin=older_clip[0], vmax=older_clip[1])
            
        if show_horizons_type is not None:
            color_iter = itertools.cycle(['lime', 'magenta', 'blue', 'cyan', 'orange', 'yellow', 'red'])
            horizons =  self.cube.get('horizons', show_horizons_type['type'])
            for name, df in horizons.items():
                X = df['X'].values.reshape(self.shape[:2])
                Y = df['Y'].values.reshape(self.shape[:2])
                Z = df['Z'].values.reshape(self.shape[:2])
                ax.plot_surface(X, Y, Z, color=next(color_iter), linewidth=0, antialiased=True)

        # overlay wells    
        if show_wells_type is not None:
            well_cmap = self.config.get_cmap(show_wells_type['cmap'])
            well_width = show_wells_type['width'] or 4
            
            all_log_df = []
            for well in self.cube.data['wells'].values():
                all_log_df.append(well.get('log'))
            combined_log = np.concatenate(all_log_df)
            older_clip = self._get_clip_values(combined_log, show_wells_type['clip'])
            
            norm_well = plt.Normalize(older_clip[0], older_clip[1])
            sm = ScalarMappable(norm=norm_well, cmap=well_cmap)
            sm.set_array(combined_log)
            
            for well_name, well in self.cube.data['wells'].items():
                well_il, well_xl = well['coord']
                # np.linspace(0, 1000, 500)
                Z = np.linspace(self.size[5], self.size[4], len(well['log']))
                ax.scatter([well_il]*len(Z), [well_xl]*len(Z), Z, 
                           c=sm.to_rgba(well['log']), s=well_width, depthshade=False)
        

        # Set viewing angle
        ax.view_init(elev=view_pos[0], azim=view_pos[1])

        # Configure axes
        ax.set_xlabel(labels[0], **self.config.label_style)
        ax.set_ylabel(labels[1], **self.config.label_style)
        ax.set_zlabel(labels[2], **self.config.label_style)

        # Set axis limits (inverted for seismic convention except xline)
        ax.set_xlim(self.size[1], self.size[0])
        ax.set_ylim(self.size[2], self.size[3])
        ax.set_zlim(self.size[4], self.size[5])

        # Enhanced visual styling
        ax.xaxis.pane.set_visible(False)
        ax.yaxis.pane.set_visible(False)
        ax.zaxis.pane.set_visible(False)
        ax.grid(False)
        # ax.grid(True, alpha=self.config.grid_alpha)

        # Apply lighting effects
        if lighting:
            ax.xaxis.pane.fill = False
            ax.yaxis.pane.fill = False
            ax.zaxis.pane.fill = False

        # Set title with proper spacing
        if title:
            ax.set_title(title, **self.config.title_style, pad=20)

        # Add colorbar
        norm = Normalize(vmin=older_clip[0], vmax=older_clip[1])
        sm = ScalarMappable(norm=norm, cmap=cmap)
        sm.set_array([])
        cbar = fig.colorbar(sm, ax=ax,
                          aspect=self.config.colorbar_aspect,
                          shrink=self.config.colorbar_shrink,
                          pad=self.config.colorbar_pad)
        if unit_label is not None:
            cbar.set_label(unit_label, **self.config.label_style, labelpad=2)
        cbar.ax.tick_params(labelsize=self.config.tick_labelsize)

        # ax.set_axis_off()

        plt.tight_layout()
        plt.show()
        

        # Save figure if requested
        if save_path:
            plt.savefig(save_path, dpi=self.config.figure_dpi,
                       bbox_inches='tight', facecolor=fig.get_facecolor(),
                       edgecolor='none')






# Copyright (c) 2025 Qi Pang.
# All rights reserved.

"""
seisvis - A seismic and geophysical visualization library

Author:
    Qi Pang

Contributor:
    Hongling Chen (valuable suggestions and feedback during development)

License:
    MIT License

Purpose:
    Developed for academic research in seismic data visualization,
    including 1D well log plotting, 2D section visualization, and 3D volumetric rendering
    of geophysical properties.

Acknowledgments:
    This project leverages open-source resources and some colormaps from:
        - Matplotlib (https://matplotlib.org/)
        - OpendtectColormaps (https://github.com/whimian/OpendtectColormaps)
    
    We gratefully acknowledge the contributors of these libraries for their valuable tools
    and visual assets, which enhance the clarity and quality of seismic interpretation.
"""

# seisvis/__init__.py

from .data_cube import DataCube
from .plot_config import PlotConfig
from .plot1d import Seis1DPlotter
from .plot2d import Seis2DPlotter
from .plot3d import Seis3DPlotter


__all__ = [
    "DataCube",
    "PlotConfig",
    "Seis1DPlotter",
    "Seis2DPlotter",
    "Seis3DPlotter",
]



# Copyright (c) 2025 Qi Pang.
# GeoAI-INV, Xi'an Jiaotong University (XJTU).
# All rights reserved.
import matplotlib
from .opendtect_colormaps import OpendtectColormaps

class PlotConfig:   
    """Configuration class for plotting, providing centralized style management"""
    
    def __init__(self):
        # Font configuration
        self.font_family = 'Times New Roman'
        self.font_weight = 'bold'
        self.tick_labelsize = 14
        self.label_fontsize = 14
        self.title_fontsize = 12
        self.legend_fontsize = 10
        self.figure_dpi = 200
        
        # Color configuration
        self.default_cmap = 'seismic'
        self.background_color = 'white'
        self.grid_alpha = 0
        
        # 3D view configuration
        self.default_elevation = 25
        self.default_azimuth = 55
        
        # Colorbar configuration
        self.colorbar_aspect = 30
        self.colorbar_shrink = 0.6
        self.colorbar_pad = 0.01
        
        # self._apply_matplotlib_settings()
    
    def apply_matplotlib_settings(self):
        """Apply global matplotlib settings"""
        matplotlib.rcParams['font.family'] = self.font_family
        matplotlib.rcParams['font.weight'] = self.font_weight
        matplotlib.rcParams['xtick.labelsize'] = self.tick_labelsize
        matplotlib.rcParams['ytick.labelsize'] = self.tick_labelsize
        matplotlib.rcParams['figure.dpi'] = self.figure_dpi
        
    def get_cmap(self, name="Petrel"):
        try:
            return matplotlib.pyplot.get_cmap(name)
        except ValueError:
            od_cmaps = OpendtectColormaps()
            return od_cmaps(name)
    
    def cmaps(self):
        od_cmaps = OpendtectColormaps()
        return od_cmaps.print_cmap_names()
    
    @property
    def label_style(self):
        return {
            "fontsize": self.label_fontsize,
            "fontweight": self.font_weight,
            "family": self.font_family
        }
    
    @property
    def title_style(self):
        return {
            "fontsize": self.title_fontsize,
            "fontweight": self.font_weight,
            "family": self.font_family
        }

    @property
    def legend_style(self):
        return {
            "fontsize": self.legend_fontsize,
            "family": self.font_family
        }


# Copyright (c) 2025 Qi Pang.
# All rights reserved.
import matplotlib.pyplot as plt
from typing import Optional, Union, Tuple, List, Dict
import numpy as np
import itertools
from scipy.interpolate import interp1d

from .plot_config import PlotConfig
from .data_cube import DataCube

ClipType = Union[str, Tuple[float, float], None]
SeismicType = Dict[str, Union[str, ClipType]]
#  {'type': 'deg10','cmap': 'gray','clip': None}  # or 'rubust' or (min, max)

class Seis2DPlotter:
    """2D data plotter"""
    
    def __init__(self, data_cube: Union[DataCube],size=None,
                       config: Optional[PlotConfig] = None):
        """
        size: [il_start, il_end, xl_start, xl_end, t_end, t_start]
        """
        self.config = config or PlotConfig()
        self.config.apply_matplotlib_settings()
        
        self.cube = data_cube
        self.size = size 
        
    @property
    def shape(self):
        if self.size is None:
            raise ValueError("shape requires 3D data. 'size' is not defined.")
        return [
            self.size[1] - self.size[0] + 1,
            self.size[3] - self.size[2] + 1,
            self.size[4] - self.size[5] + 1,
        ]

    @property
    def extent_inline_2d(self):
        if self.size is None:
            raise ValueError("extent_inline_2d requires 3D data. 'size' is not defined.")
        return [self.size[2], self.size[3], self.size[4], self.size[5]]

    @property
    def extent_xline_2d(self):
        if self.size is None:
            raise ValueError("extent_xline_2d requires 3D data. 'size' is not defined.")
        return [self.size[0], self.size[1], self.size[4], self.size[5]]

    @property
    def extent_surface(self):
        if self.size is None:
            raise ValueError("extent_surface requires 3D data. 'size' is not defined.")
        return [self.size[2], self.size[3], self.size[0], self.size[1]]

        
    def _get_clip_values(self, array: np.ndarray, 
                        clip: Optional[Union[Tuple[float, float], str]] = None) -> Tuple[float, float]:
        "Calculate data clipping range with various methods"
        if clip is None:
            return array.min(), array.max()
        elif isinstance(clip, str):
            if clip == 'robust':
                # Robust clipping using IQR
                q25, q75 = np.percentile(array, [25, 75])
                iqr = q75 - q25
                return q25 - 1.5*iqr, q75 + 1.5*iqr                
            else:
                raise ValueError(f"Unsupported clip type: {clip}")
        else:
            return float(clip[0]), float(clip[1])
        
    def _make_surface(self, raw_surface):
        
        surface_shape = self.shape[:2]
        full_surface = np.full(surface_shape, np.nan, dtype=np.float32)
        ils = np.arange(self.size[0], self.size[1]+1, 1, dtype=int)
        xls = np.arange(self.size[2], self.size[3]+1, 1, dtype=int)
        
        for idx, inline in enumerate(ils):
            x = raw_surface.loc[raw_surface.X == inline].Y 
            y = raw_surface.loc[raw_surface.X == inline].Z
            interp = interp1d(x, y, bounds_error=False, fill_value='extrapolate')
            full_surface[idx, :] = interp(xls) # (inline, crossline)
            
        return full_surface
    
    def _calculate_mask_from_horizons(self, top_h, base_h, dt=1):
        top_s = self._make_surface(top_h)
        base_s = self._make_surface(base_h)
        mask = np.zeros(self.shape, dtype=bool)  
    
        for i in range(mask.shape[0]):
            tss = (top_s[i, :] - self.size[5]) // dt
            mas = (base_s[i, :] - self.size[5]) // dt
            for j in range(len(tss)):
                start = int(tss[j])
                end = int(mas[j])
                if start < end:
                    mask[i, j, start:end] = True
                else:
                    mask[i, j, end:start] = True
        return mask
        
    def plot_surface(self, surface: np.ndarray, 
                    figsize: Tuple[int, int] = (7, 5),
                    cmap: str = 'viridis',
                    clip: Optional[Union[Tuple[float, float], str]] = None,
                    labels: Optional[Tuple[str, str]] = ('X', 'Y'),
                    title: Optional[str] = None,
                    well_positions: Optional[np.ndarray] = None,
                    well_values = None, 
                    contour_levels: int = 40) -> None:
 
        if surface.ndim != 2:
            raise ValueError("Surface data must be 2D")
        
        fig, ax = plt.subplots(figsize=figsize)
        
        vmin, vmax = self._get_clip_values(surface, clip)
        

        c = ax.imshow(surface, aspect='auto', cmap=self.config.get_cmap(cmap),
                      vmin=vmin, vmax=vmax, extent=self.extent_surface)
        
        # ax.contour(surface, colors='k', levels=contour_levels, linewidths=0.5,
        #           extent=extent, origin='upper')        

        if well_positions is not None:
            inline_coords = well_positions[:, 0]
            xline_coords = well_positions[:, 1]
            ax.scatter(xline_coords, inline_coords, c=well_values, 
                       cmap=self.config.get_cmap(cmap),vmin=vmin, vmax=vmax, 
                       s=100, marker='o', 
                       edgecolors='white', linewidths=1)
            # ax.legend()        
            # ax.grid(True)
        fig.colorbar(c, ax=ax, aspect=30, shrink=1.0, pad=0.01)
        
        if labels is not None:
            ax.set_xlabel(labels[0], **self.config.label_style)
            ax.set_ylabel(labels[1], **self.config.label_style)
        
        if title:
            ax.set_title(title, **self.config.title_style)
        
        plt.tight_layout()
        plt.show()
                
        
    def _extract_slice(self, section_idx, section_type, il_start, xl_start):
        
        if section_type == 'inline':
            assert il_start is not None, "You must provide il_start for inline section."
            il_idx = section_idx - il_start
            slice_2d_func = lambda cube: cube[il_idx, :, :].T
            horizon_filter = lambda df: df.loc[df.X == section_idx]
            x_label = "Crossline"
            z_axis_label = None
            title = f"Inline {section_idx}"
            extent = self.extent_inline_2d
        elif section_type == 'xline':
            assert xl_start is not None, "You must provide xl_start for crossline section."
            xl_idx = section_idx - xl_start
            slice_2d_func = lambda cube: cube[:, xl_idx, :].T
            horizon_filter = lambda df: df.loc[df.Y == section_idx]
            x_label = "Inline"
            z_axis_label = None
            title = f"Crossline {section_idx}"
            extent = self.extent_xline_2d
        elif section_type == 'time':
            t_idx = section_idx
            slice_2d_func = lambda cube: cube[:, :, t_idx]
            horizon_filter = None  
            x_label = "Crossline"
            z_axis_label="Inline"
            title = f"Time Slice {section_idx}"
            extent = self.extent_surface  
    
        else:
            raise ValueError("section_type must be 'inline', 'xline', or 'time'")
        
        return slice_2d_func, horizon_filter, x_label, title, extent, z_axis_label    
               
    def plot_section(self, 
        section_idx=None,
        section_type='inline',
        show_seismic_type: Optional[SeismicType] = None,
        show_properties_type: Optional[SeismicType] = None,
        show_horizons_type: Optional[SeismicType] = None,
        show_wells_type: Optional[SeismicType] = None,
        t_axis_label="Time (ms)",
        unit_label_p = None, # 'g/cm³·m/s'
        figsize: Tuple[int, int] = (10, 4),
        save_path=None,  
        **kwargs
    ):  
        "Plot a 2D section (slice) from a 3D seismic or property cube."
        fig, ax = plt.subplots(figsize=figsize)
        
        slice_2d_func, horizon_filter, x_label, title, extent, z_axis_label = self._extract_slice(section_idx, section_type, 
                                                                            self.size[0], self.size[2])
        if z_axis_label is None:
            z_axis_label = t_axis_label
        
        if show_seismic_type is not None:
            seis = self.cube.get('seismic', show_seismic_type['type'])
            if seis.ndim != 3:
                raise ValueError(f"plot_section expects 3D data, got {seis.ndim}D for '{show_seismic_type['type']}'.\n"
                                 f"If you only have 2D data, please use plot_2d_section instead.")
            seis_clip = self._get_clip_values(seis, show_seismic_type['clip'])
            
            if show_seismic_type['mask'] and show_horizons_type is not None  \
            and len(show_horizons_type['type'])>1:
                # print('yes')
                top_h =  self.cube.get('horizons', show_horizons_type['type'][0])
                base_h = self.cube.get('horizons', show_horizons_type['type'][-1])
                mask = self._calculate_mask_from_horizons(top_h, base_h)
                seis = seis * mask
            
            seis_sliced = slice_2d_func(seis)
            cmap = self.config.get_cmap(show_seismic_type['cmap']).copy()
            cmap.set_bad('white', alpha=0)
            
            seis_sliced = slice_2d_func(seis)
            c1 = ax.imshow(seis_sliced, aspect='auto', cmap=cmap,
                      vmin=seis_clip[0], vmax=seis_clip[1],
                      extent=extent, origin='upper', **kwargs)
            
        if show_properties_type is not None:
            
            prop = self.cube.get('properties', show_properties_type['type'])   
            if prop.ndim != 3:
                raise ValueError(f"plot_section expects 3D data, got {seis.ndim}D for '{show_properties_type['type']}'.\n"
                                 f"If you only have 2D data, please use plot_2d_section instead.")
            prop_clip = self._get_clip_values(prop, show_properties_type['clip'])  
            
            if show_properties_type['mask'] and show_horizons_type is not None \
            and len(show_horizons_type['type'])>1:
                # print('yes')
                top_h =  self.cube.get('horizons', show_horizons_type['type'][0])
                base_h = self.cube.get('horizons', show_horizons_type['type'][-1])
                mask = self._calculate_mask_from_horizons(top_h, base_h)
                prop = np.ma.array(prop, mask=~mask)
                
                
            prop_sliced = slice_2d_func(prop)
            # Note:
            # Some issues may occur when using OpenDtect-style colormaps with masked data.
            # In particular, matplotlib may not handle 'set_bad()' correctly with certain custom colormaps,
            # resulting in masked (e.g., NaN) regions not being shown as transparent or white.
            # use Ture or False
            cmap = self.config.get_cmap(show_properties_type['cmap']).copy()
            cmap.set_bad('white', alpha=0)
            
            c2 = ax.imshow(prop_sliced, aspect='auto', cmap=cmap,
                      vmin=prop_clip[0], vmax=prop_clip[1],
                      extent=extent, origin='upper', **kwargs)
            
            # ax.contour(
            # prop_sliced[::-1, :], extent=extent, levels=30,
            # colors='k', linewidths=0.6)
            
        if show_horizons_type is not None:
            hor = self.cube.get('horizons', show_horizons_type['type'])
            color_iter = itertools.cycle(['lime', 'magenta', 'blue', 'cyan', 'orange', 'yellow', 'red'])
            for name, df in hor.items():
                df_filtered = horizon_filter(df)
                if not df_filtered.empty:
                    x_vals = df_filtered.Y if section_type == 'inline' else df_filtered.X
                    z_vals = df_filtered.Z
                    ax.plot(x_vals, z_vals, lw=2.5, label=name, color=next(color_iter))
                    # ax.legend(fontsize=self.config.legend_fontsize)
                
        if show_wells_type is not None:
            well_cmap = show_wells_type['cmap']
            well_width = show_wells_type['width'] or 4
            all_log_df = []
            for well in self.cube.data['wells'].values():
                all_log_df.append(well.get('log'))
            combined_log = np.concatenate(all_log_df)
            # different clip is better? because different prop
            if show_properties_type is not None: 
                well_clip = prop_clip
            else:
                well_clip = self._get_clip_values(combined_log, show_wells_type['clip'])
            for well_name, well in self.cube.data['wells'].items():
                well_il, well_xl = well['coord']
                if section_type == 'inline': 
                    if well_il == section_idx:
                        ax.imshow(well['log'], aspect='auto', cmap=self.config.get_cmap(well_cmap), 
                                  vmin=well_clip[0], vmax=well_clip[1],
                                  extent=(well_xl - well_width//2, well_xl + well_width//2, self.size[4], self.size[5]), 
                                  zorder=20)
                else:
                    if well_xl == section_idx:
                        ax.imshow(well['log'], aspect='auto', cmap=self.config.get_cmap(well_cmap), 
                                  vmin=well_clip[0], vmax=well_clip[1],
                                  extent=(well_il - well_width//2, well_il + well_width//2, self.size[4], self.size[5]), 
                                  zorder=20)
        
        ax.set_xlim(extent[0], extent[1])
        ax.set_ylim(extent[2], extent[3])
        
        # Labels & title
        ax.set_xlabel(x_label, **self.config.label_style)
        ax.set_ylabel(z_axis_label, **self.config.label_style)
        ax.set_title(title, **self.config.title_style)
        
  
        if show_seismic_type is not None and show_seismic_type['bar']:
            cbar = fig.colorbar(c1, ax=ax, aspect=30, shrink=1.0, pad=0.01)
            # cbar.set_label(unit_label, **self.config.label_style, labelpad=2)
        elif show_properties_type is not None and show_properties_type['bar']:
            cbar = fig.colorbar(c2, ax=ax, aspect=30, shrink=1.0, pad=0.01)
            if unit_label_p is not None:
                cbar.set_label(unit_label_p, **self.config.label_style, labelpad=2)
        
        plt.tight_layout()
        
        if save_path:
            plt.savefig(save_path, dpi=self.config.figure_dpi, bbox_inches='tight')
        
        plt.show()
        
        
    def plot_2d_section(self,
        show_seismic_type: Optional[SeismicType] = None,
        show_properties_type: Optional[SeismicType] = None,
        show_wells_type: Optional[SeismicType] = None,
        y_axis_label="Time (ms)",
        x_label="Trace",
        figsize: Tuple[int, int] = (10, 4),
        extent=None,
        show_colorbar=True,
        save_path=None,
        **kwargs
    ):
        "Plot a 2D section from 2D seismic or property data."
        show_type = show_seismic_type or show_properties_type
        if show_seismic_type is not None:
            category = 'seismic'
        elif show_properties_type is not None:
            category = 'properties'
        else:
            raise ValueError("Unknown data")
            
        dim = self.cube.get_dim(category, show_type['type'])
        
        if dim != 2:
            raise ValueError(f"plot_2d_section expects 2D data, got {dim}D for "
                             f"'{show_type['category']}':'{show_type['type']}'.\n"
                             "If you want to plot a slice from 3D data, please use plot_section instead.")
        data_2d = self.cube.get(category, show_type['type'])
        # Setup figure
        fig, ax = plt.subplots(figsize=figsize)
        
        # Clipping
        clip = self._get_clip_values(data_2d, show_type['clip'])
        
        # extent = data_2d.shape
        
        if extent is None:
            extent = [0, data_2d.shape[1], data_2d.shape[0], 0]
        
        # Main section plot
        c = ax.imshow(
            data_2d, aspect='auto', cmap=self.config.get_cmap(show_type['cmap']),
            vmin=clip[0], vmax=clip[1],
            extent=extent, origin='upper', **kwargs
        )
                
        # Wells overlay
        if show_wells_type is not None:  
            well_cmap = show_wells_type['cmap']
            well_width = show_wells_type['width'] or 4
            all_log_df = []
            for well in self.cube.data['wells'].values():
                all_log_df.append(well.get('log'))
            combined_log = np.concatenate(all_log_df)
            # different clip is better? because different prop
            if show_properties_type is not None: 
                well_clip = clip
            else:
                well_clip = self._get_clip_values(combined_log, show_wells_type['clip'])
            
            for well_name, well in self.cube.data['wells'].items():
                pos = well['coord']
                ax.imshow(well['log'], aspect='auto', cmap=self.config.get_cmap(well_cmap), 
                          vmin=well_clip[0], vmax=well_clip[1],
                          extent=(pos - well_width//2, pos + well_width//2, extent[-2], extent[-1]), zorder=20)
        
        # Colorbar
        if show_colorbar:
            cbar = fig.colorbar(c, ax=ax, aspect=30, shrink=1.0, pad=0.01)
            # cbar.ax.text(0.5, 1.02, 'g/cm³·m/s', ha='center', va='bottom', fontsize=self.config.label_fontsize, transform=cbar.ax.transAxes)
            # cbar.set_label('(g/cm³)', **self.config.label_style, labelpad=2)
            if show_type['unit_label'] is not None:
                cbar.set_label(show_type['unit_label'], **self.config.label_style, labelpad=2)
        
        # # there is no better way to plot only 2d horizon
        # if horizons is not None:
        #     color_iter = itertools.cycle(['lime', 'magenta', 'blue', 'cyan', 'orange', 'yellow', 'red'])
        #     for name, df in horizons.items():
        #         if not df.empty:
        #             x_vals = df.X if 'X' in df.columns else df.index
        #             z_vals = df.Z if 'Z' in df.columns else df.values
        #             ax.plot(x_vals, z_vals, lw=2.5, label=name, color=next(color_iter))
        #     ax.legend(loc='upper right')
        
        ax.set_xlim(extent[0], extent[1])
        ax.set_ylim(extent[2], extent[3])
        
        # Labels & title
        ax.set_xlabel(x_label, **self.config.label_style)
        ax.set_ylabel(y_axis_label, **self.config.label_style)
        # ax.set_title(title, **self.config.title_style)
        
        plt.tight_layout()
        
        if save_path:
            plt.savefig(save_path, dpi=self.config.figure_dpi, bbox_inches='tight')
        
        plt.show()
        
    def plot_2d_section_plus(self,
        show_seismic_type: Optional[SeismicType] = None,
        show_properties_type: Optional[Union[SeismicType, list]] = None,  
        show_wells_type: Optional[SeismicType] = None,
        y_axis_label="Time (ms)",
        x_label="Trace",
        show_labels=True,
        figsize: Tuple[int, int] = (10, 4),
        extent=None,
        show_colorbar=True,
        ncols: int = 0, 
        save_path=None,
        **kwargs
    ):
        """
        Plot one or multiple 2D property/seismic sections.
        If `show_properties_type` is a list, multiple panels will be plotted in one figure.
        """
        # === Determine data category ===
        show_type = show_seismic_type or show_properties_type
        if show_seismic_type is not None:
            category = 'seismic'
        elif show_properties_type is not None:
            category = 'properties'
        else:
            raise ValueError("Unknown data")
    
        # === Handle multiple properties ===
        if isinstance(show_properties_type, list):
            prop_list = show_properties_type
            n = len(prop_list)
            if ncols <= 0:
                ncols = int(np.ceil(np.sqrt(n)))
            nrows = int(np.ceil(n / ncols))
            fig, axes = plt.subplots(nrows, ncols, figsize=(figsize[0]*ncols, figsize[1]*nrows))
            axes = np.atleast_2d(axes)
    
            for i, show_type in enumerate(prop_list):
                r, c = divmod(i, ncols)
                ax = axes[r, c]
    
                dim = self.cube.get_dim(category, show_type['type'])
                if dim != 2:
                    raise ValueError(f"plot_2d_section expects 2D data, got {dim}D for "
                                     f"'{category}:{show_type['type']}'.")
    
                data_2d = self.cube.get(category, show_type['type'])
                clip = self._get_clip_values(data_2d, show_type['clip'])
                if extent is None:
                    extent_local = [0, data_2d.shape[1], data_2d.shape[0], 0]
                else:
                    extent_local = extent
    
                cax = ax.imshow(
                    data_2d, aspect='auto', cmap=self.config.get_cmap(show_type['cmap']),
                    vmin=clip[0], vmax=clip[1], extent=extent_local, origin='upper', **kwargs
                )
                ax.set_xlim(extent_local[0], extent_local[1])
                ax.set_ylim(extent_local[2], extent_local[3])
                if show_labels:
                    ax.set_title(show_type['type'], **self.config.label_style)
                    ax.set_xlabel(x_label, **self.config.label_style)
                    ax.set_ylabel(y_axis_label, **self.config.label_style)
                else:
                    ax.set_xticks([])
                    ax.set_yticks([])   
                    
                if show_colorbar:
                    fig.colorbar(cax, ax=ax, aspect=30, shrink=1.0, pad=0.01)
    
            # Remove empty subplots
            for j in range(i+1, nrows*ncols):
                r, c = divmod(j, ncols)
                axes[r, c].axis('off')
    
            plt.tight_layout()
            if save_path:
                plt.savefig(save_path, dpi=self.config.figure_dpi, bbox_inches='tight')
            plt.show()
            return
    
        # === Single property/seismic plot===
        dim = self.cube.get_dim(category, show_type['type'])
        if dim != 2:
            raise ValueError(f"plot_2d_section expects 2D data, got {dim}D for "
                             f"'{show_type['category']}':'{show_type['type']}'.")
    
        data_2d = self.cube.get(category, show_type['type'])
        fig, ax = plt.subplots(figsize=figsize)
        clip = self._get_clip_values(data_2d, show_type['clip'])
    
        if extent is None:
            extent = [0, data_2d.shape[1], data_2d.shape[0], 0]
    
        c = ax.imshow(
            data_2d, aspect='auto', cmap=self.config.get_cmap(show_type['cmap']),
            vmin=clip[0], vmax=clip[1],
            extent=extent, origin='upper', **kwargs
        )
    
        if show_colorbar:
            cbar = fig.colorbar(c, ax=ax, aspect=30, shrink=1.0, pad=0.01)
            if show_type['unit_label'] is not None:
                cbar.set_label(show_type['unit_label'], **self.config.label_style, labelpad=2)
    
        ax.set_xlim(extent[0], extent[1])
        ax.set_ylim(extent[2], extent[3])
        ax.set_xlabel(x_label, **self.config.label_style)
        ax.set_ylabel(y_axis_label, **self.config.label_style)
        plt.tight_layout()
    
        if save_path:
            plt.savefig(save_path, dpi=self.config.figure_dpi, bbox_inches='tight')
        plt.show()





# Copyright (c) 2025 Qi Pang.
# All rights reserved.

import numpy as np

class DataCube:
    def __init__(self):
        self.data = {
            'seismic': {},     # seismic data, eg: {'10 degree': ndarray, '20 degree': ndarray}
            'properties': {},  # properties, eg: {'PHIE': ndarray, 'VCL': ndarray}
            'horizons': {},    # horizons, eg: {'Top': ndarray}
            'wells': {}        # wells, eg: {'Well-A': {'log': ndarray, 'coord': (inline, xline)}}
        }

    def add_seismic(self, name, data):
        self.data['seismic'][name] = data

    def add_property(self, name, data):
        self.data['properties'][name] = data

    def add_horizon(self, name, data):
        self.data['horizons'][name] = data
    
    # need to be updated to {'Well-A': {'log': df(all properties not one), 'coord': (inline, xline)}}
    def add_well(self, well_name, data):  
        self.data['wells'][well_name] = data
    
    def remove(self, category, name):
        if category in self.data:
            self.data[category].pop(name, None)

    def get(self, category, name=None):
        category_data = self.data.get(category, {}) 
        if name is None:
            return category_data
        if isinstance(name, list):
            return {n: category_data.get(n, None) for n in name}
        # name is a single string
        return category_data.get(name, None)
    
    def get_dim(self, category, name):
        arr = self.get(category, name)
        if isinstance(arr, np.ndarray):
            return arr.ndim
        return None
    
    def summary(self):
        print("DataCube Information:")
        for category, content in self.data.items():
            print(f"- {category}: {list(content.keys())}")
            
    def clear(self):
        for key in self.data:
            self.data[key].clear()



# Copyright (c) 2025 Qi Pang.
# All rights reserved.
import matplotlib.pyplot as plt
import numpy as np
from typing import Optional, Tuple, List
import itertools
from .plot_config import PlotConfig

class Seis1DPlotter:
    """1D data plotter"""
    
    def __init__(self, config: Optional[PlotConfig] = None):
        self.config = config or PlotConfig()
        self.config.apply_matplotlib_settings()

    def plot_groups(self,
                    data_groups: List[List[np.ndarray]],
                    t_start=None,
                    dt = 1,
                    titles: Optional[List[str]] = None,
                    legends: Optional[List[str]] = None,
                    line_styles: Optional[List[str]] = None,
                    vis_type='v',
                    figsize: Tuple[int, int] = (4, 11),
                    save_path: Optional[str] = None):
        """
        Plot each data group (e.g., each variable) in a subplot, all in one figure.
        """
        if isinstance(data_groups, np.ndarray):
            data_groups = [[data_groups]]
        elif isinstance(data_groups, list) and isinstance(data_groups[0], np.ndarray):
            data_groups = [data_groups]
            
        if isinstance(titles, str):
            titles = [titles]
            
        if isinstance(legends, str):
            legends = [legends]    
            
        n_groups = len(data_groups)
        n_traces = len(data_groups[0])
        
        # --- Check all groups have the same number of traces ---
        for g in data_groups:
            if len(g) != n_traces:
                raise ValueError("All groups must have the same number of traces (methods).")
        
        # --- Check titles ---
        if titles is not None:
            if len(titles) != n_groups:
                raise ValueError(f"Number of titles {len(titles)} must match number of groups {n_groups}.")
                
        # --- Check legends ---
        if legends is not None:
            if len(legends) != n_traces:
                raise ValueError(f"Number of legends {len(legends)} must match number of traces per group {n_traces}.")
        
        if t_start is None:
            time = (np.arange(data_groups[0][0].shape[0])) * dt
        else:
            time = (np.arange(data_groups[0][0].shape[0])) * dt + t_start
    
        if line_styles is None:
            line_styles = ['-'] * n_traces
    
        color_iter = itertools.cycle([
            'black', 'blue', 'red', 'green', 'orange',
            'purple', 'cyan', 'magenta', 'brown', 'olive'
        ])
        color_list = list(itertools.islice(color_iter, n_traces))
        
        if vis_type == 'v':
            fig, axes = plt.subplots(1, n_groups, figsize=(figsize[0]*n_groups, figsize[1]), sharey=True)
        else:
            fig, axes = plt.subplots(n_groups, 1, figsize=(figsize[0], figsize[1]*n_groups), sharey=True)
            
        if n_groups == 1:
            axes = [axes]
    
        for i, traces in enumerate(data_groups):
            ax = axes[i]
            for j, trace in enumerate(traces):
                label = legends[j] if legends else f'Trace {j+1}'
                if vis_type == 'v':
                    ax.plot(trace, time, line_styles[j], color=color_list[j], label=label)
                    y_axis_label = "Time (ms)"
                    ax.set_xlabel("Amplitude", **self.config.label_style)
                else:
                    ax.plot(time, trace, line_styles[j], color=color_list[j], label=label)
                    x_axis_label = "Time (ms)"
                    ax.set_ylabel("Amplitude", **self.config.label_style)

            if titles:
                ax.set_title(titles[i], **self.config.title_style)
                
            if vis_type == 'v':    
                ax.margins(x=0.05, y=0)
            else:
                ax.margins(x=0, y=0.05)
            
            if vis_type == 'v':
                ax.invert_yaxis()
                
            if i == 0 and vis_type == 'v':
                ax.set_ylabel(y_axis_label, **self.config.label_style)
                
            if i == n_groups - 1 and vis_type != 'v':
                ax.set_xlabel(x_axis_label, **self.config.label_style)
        
            ax.legend(loc='upper right', fontsize=self.config.legend_fontsize)
    
        plt.tight_layout()
        if save_path:
            plt.savefig(save_path, dpi=self.config.figure_dpi)
        plt.show()

